#-------------------------------------------------------------------------------
# ChainLaddere Constructor
#-------------------------------------------------------------------------------
ChainLadder <- function(triangle, DFs, ...)
{
  # Check triangle is of correct class
  if(!("Triangle" %in% class(triangle)))
    stop("triangle should be of class: Triangle")

  # Get ultimate development factors.  Can pass arguments to
  # calculateAvDFs function.
  if(missing(DFs))
    DFs <- calculateUltDFs(triangle, ...)

  # Get ultimates
  .ult <- calculateCLUlts(triangle, DFs)

  # Get expected claims
  .exp <- makeIncremental(calculateCLExpClaims(.ult, DFs))

  # Output results
  .out <- list( call = match.call(),
                ult = .ult,
                data = triangle,
                DFs = DFs,
                exp = .exp)
                
  attr(.out, "class") <- c("ChainLadder", "Projection", "list")

  return(.out)
}



#-------------------------------------------------------------------------------
# Just calculate ultimates (for Bootstrapping, would be quicker)
#-------------------------------------------------------------------------------
calculateCLUlts <- function(triangle, DFs)
{
  # Get latest claims data
  if(triangle$cumulative==F) triangle <- makeCumulative(triangle)
  .latest <- getLatest(triangle)

  # Match development factors to latest data
  .DFs <- .getLatestDFs(DFs, .latest)

  # Multiply to get ultimates
  .ult <- .latest$latest * .DFs

  return(.ult)
}


#-------------------------------------------------------------------------------
# Get expected claims
#-------------------------------------------------------------------------------
calculateCLExpClaims <- function(ult, DFs)
{
  .ult <- .expandArray(ult, 2, dim(DFs$ultDFs)[2])
  .ultDFs <- .expandArray(DFs$ultDFs, 1, dim(ult)[1])
  .exp <- .ult / .ultDFs

  .dimnames <- dimnames(ult)
  .dimnames$dev <- dimnames(DFs$ultDFs)$dev
  dimnames(.exp)<- .dimnames
  
  .exp <- ArrayTriangle(.exp)
  
  return(.exp)
}



#-------------------------------------------------------------------------------
# print Method
#-------------------------------------------------------------------------------
print.ChainLadder <- function(x, triangle=1)
{
  .latest <- getLatest(makeCumulative(x$data))
  .DFs <- .getLatestDFs(x$DFs, .latest)


  .out <- data.frame( Latest = .latest$latest[,, triangle],
                      CLDev = 100 * rev(1 / .DFs[,, triangle]),
                      Reserve = x$ult[,, triangle], #Dummy
                      Ult = x$ult[,, triangle]
                    )
  .out$Reserve <- .out$Ult - .out$Latest
  
  .totals <- unlist(lapply(.out, sum))
  .totals[2] <- NA
  .out <- rbind(.out, .totals)
  rownames(.out)[dim(.out)[1]] <- "Totals"
  print(format(.out,  big.mark = ",", digits = 3 ))
  
}

