# ******************************************************************************
# Class Methods
# ******************************************************************************

#-------------------------------------------------------------------------------
# Print methods
#-------------------------------------------------------------------------------
print.DFs <- function(x)
{
  print(x$ultDFs)
}


#-------------------------------------------------------------------------------
# Plot method for ArrayTriangle
#-------------------------------------------------------------------------------
plot.DFs <- function(x, triangle=1, what="AvDFs", plotTail=F, ...)
{  
  # Limits of graph
  if(is.null(x$curveDFs) | plotTail==F) .xlim <- c(0, dim(x$avDFs)[2])
  else .xlim <- c(0, dim(x$curveDFs)[2])
    
  if(what=="AvDFs"){
    # Convert Triangle of DFs to Long format
    .indivDFs <- .formatLong(x$indivDFs, na.rm=T)
    .indivDFs <- .indivDFs[.indivDFs$triangle==triangle, ]

    # Careful as factors - MAY NEED TO REVISIT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    .x <- as.numeric(.indivDFs$dev)
    plot(.x, .indivDFs$data, pch="x", cex=0.8, xlim=.xlim,
      xlab="Develpoment Period", ylab="Development Factor",
      main="Plot of Development Factors", ...)

    # Careful as factors - MAY NEED TO REVISIT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    .x <- as.numeric(dimnames(x$avDFs)$dev)
    lines(.x, x$avDFs[1, , triangle], col="red")

    # Careful as factors - MAY NEED TO REVISIT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if(!is.null(x$curveDFs)) {
      .x <- as.numeric(dimnames(x$curveDFs)$dev)
      lines(.x, x$curveDFs[1, , triangle], col="green")
    }
  }

  if(what=="PercDev"){
    # Careful as factors - MAY NEED TO REVISIT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    .x <- as.numeric(dimnames(x$ultDFs)$dev)
    .y <- 1 / x$ultDFs[1, , triangle]
    plot(.x, .y, type="l", xlim=.xlim,
      xlab="Develpoment Period", ylab="Percentage Developed",
      main="Percentage Developed", ...)

  }

}



#-------------------------------------------------------------------------------
# Function to calculate ultimate development factors
#-------------------------------------------------------------------------------
calculateUltDFs <- function(triangle, wghts, include, fitCurve=F, 
  useCurveFrom, ...)
{
  # Calculate individual Development Factors
  triangle <- makeCumulative(triangle)
  .indivDFs <- .getIndivDFs(triangle)

  # Determine weights
  if(missing(wghts))
    wghts <- as.ArrayTriangle(triangle)$data
  else {
    # !!!!!!!!!!!!!!!Check appropriate format - LATER !!!!!!!!!!!!!!!!!!!!!!!!!!
  }

  # Determine factors to include in calculation
  if(missing(include)) {
    include <- .indivDFs
    include[!is.na(include)] <- 1}
  else {
    # !!!!!!!!!!!!!!!Check appropriate format - LATER !!!!!!!!!!!!!!!!!!!!!!!!!!
  }

  # Determine method
  # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  # Could have a method variable, which takes last 5 development factors, for
  # example - LATER
  # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


  # Perform calculation to get average development factor
  .avDFs <- .calculateAvDFs(.indivDFs, wghts, include)
  
  # Fit curve (to project tail)
  if(fitCurve) {
    # Fit curve
    .curveDFs <- .fitCurveDFs(.avDFs, ...)
    
    # Choose final development factors
    if(missing(useCurveFrom)) useCurveFrom <- dim(.avDFs)[2]
    .finalDFs <- .curveDFs
    .finalDFs[ , 1:(useCurveFrom-1), ] <- .avDFs[ , 1:(useCurveFrom-1), ]   
  }
  else {
    .finalDFs <- .avDFs
    .curveDFs <- NULL
  }

  # Need function to perform calculation
  .ultDF <- function(x)
  {
    return(rev(cumprod(rev(x))))
  }

  # Perform calculation
  .ultDFs <- apply(.finalDFs, c(1,3), .ultDF)
  .ultDFs <- aperm(.ultDFs, c(2,1,3))
  dimnames(.ultDFs) <- dimnames(.finalDFs)

  # Return output
  .out <- list( call = match.call(),
                indivDFs = .indivDFs,
                weights = wghts,
                avDFs = .avDFs,
                curveDFs = .curveDFs,
                finalDFs = .finalDFs,
                ultDFs = .ultDFs)
  attr(.out, "class") <- c("DFs", "list")
  return(.out)

}









#-------------------------------------------------------------------------------
# Internal: Function to get average development factors
#-------------------------------------------------------------------------------
.calculateAvDFs <- function(dfs, wghts, include)
{
  # Perform calculations
  .avDFs <- apply(dfs * wghts * include, c(2,3), sum, na.rm=T) /
    apply(wghts * include, c(2,3), sum, na.rm=T)
  .avDFs[is.nan(.avDFs)] <- 1

  .avDFs <- array(.avDFs, dim=c(1, dim(.avDFs)))

  .names <- dimnames(dfs)

  .names$origin <- 1
  dimnames(.avDFs) <- .names

  return(.avDFs)
}


#-------------------------------------------------------------------------------
# Internal: Function to calculate tail factors
#-------------------------------------------------------------------------------
.fitCurveDFs <- function(dfs, curveType="Exponential", tailLength=20)
{
  # Convert to data frame
  .avDFs <- .formatLong(dfs)
  .avDFs$dev <- as.numeric(as.character(.avDFs$dev))
  
  # Remove DFs close to and below 1 
  .avDFs$weights <- 1
  .avDFs$weights[.avDFs$data < 1.0001] <- 0
  
  # Determine length of projection
  .projectLimit <- max(.avDFs$dev) + tailLength
  
  # Define function to fit model and predict development factors
  if(curveType=="Exponential") {
  # Exponential decay model 
    .fitModel <- function(x) {
      .model <-  lm(log(data - 1) ~ dev, weights=weights, data=x) 
      .fit <- data.frame(dev=1:.projectLimit)
      .fit <- 1 + exp(predict(.model, .fit))
      return(.fit) 
    }
  }
  
  # Perform log regression
  .fitDFs <- by(.avDFs, .avDFs$triangle, .fitModel) 

  # Format output
  .fitDFs <- as.vector(unlist(.fitDFs))
  .fitDFs <- array(.fitDFs, dim=c(1, .projectLimit, dim(dfs)[3]))
  .names <- dimnames(dfs)
  .names$dev <- 1:.projectLimit
  dimnames(.fitDFs)<- .names
  
  # Return output
  return(.fitDFs)
}


#-------------------------------------------------------------------------------
# Internal: Function to get latest Development Factors
#-------------------------------------------------------------------------------
.getLatestDFs <- function(DFs, latest)
{
  # Match development factors to latest data
  .pos <- data.frame(origin=1, dev=as.numeric(as.vector(latest$dev)),
    triangle=rep(as.numeric(dimnames(latest$dev)$triangle),
    each=dim(latest$dev)[1]))
  .pos$dev <- match(.pos$dev, dimnames(DFs$ultDFs)$dev)
  .pos <- as.matrix(.pos)

  # Then pull through relevant development factors
  .DFs <- latest$latest
  .DFs[] <- DFs$ultDFs[.pos]

  # Return values
  return(.DFs)
}